#!/usr/bin/python2
#
# Copyright (c) 2013 Intel, Inc.
# Author igor.stoppa@intel.com
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; version 2 of the License
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the
# Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""Tool for handling Cleware USB Cutter devices"""

from __future__ import print_function
import re
import time
from tizentestrobot.cmdlinetool import CmdLineTool
from collections import namedtuple

CutterChannel = namedtuple("CutterChannel",
                    ["cutter_type", "cutter_id", "cutter_ch"])

class Cutter(CmdLineTool):
    """Class for handling Cleware Cutter devices"""
    def __init__(self):
        CmdLineTool.__init__(self, "clewarecontrol")
        self._channels = []
        self._opening_transient = 2
        self._closing_transient = 5

    def run(self, parms = [], cutter_id = None):
        """Execute the cutter tool with custom parameters"""
        local_parms = ["-c", "1"] #execute the command only once
        local_parms += parms
        if cutter_id: #device_id must precede commands for that device_id
            local_parms = ["-d", "{0}".format(cutter_id)] + local_parms
        return CmdLineTool.run(self, local_parms)

    def probe_channels(self, force = False):
        """Get list of connected Cutter devices"""
        if self._channels != [] and not force:
            return self._channels
        print("Detecting cutters connected ... ", end="")
        cmd_output = self.run(["-l"])
        cmd_output = cmd_output.split("\n")
        assert ("Cleware library version:" in cmd_output[0]) and \
               ("Number of Cleware devices found:" in cmd_output[1]) and\
               (cmd_output[len(cmd_output) - 1] == "")
        for i in range(2, len(cmd_output)):
            # The string has format:
            # "Device: X, type: Switch1 (Y),
            # version: Z, serial number: S/N"
            # where X, Y & Z are currently dontcare
            if "Switch1" in cmd_output[i]: #check that it's a cutter
                values = re.findall(r'\b\d+\b', cmd_output[i])
                assert len(values) == 4
                #usbcutter has version set to 5, only 1 channel
                if "5" == values[2]:
                    self._channels.append(CutterChannel(
                        cutter_type = "usb",
                        cutter_id = values[3],
                        cutter_ch = "0"
                    ))
                #mains cutter, 4 channels has version set to 29
                elif "29" == values[2]:
                    for i in range(0, 4):
                        self._channels.append(CutterChannel(
                            cutter_type = "mains",
                            cutter_id = values[3],
                            cutter_ch = "{0}".format(i)
                        ))
        print("done.")
        return self._channels

    def get_channel(self, cutter_id, cutter_ch):
        """Method returning a channel handler, given id and number"""
        if cutter_id == None or cutter_ch == None:
            return None
        self.probe_channels()
        for channel in self._channels:
            if channel.cutter_id == cutter_id and\
               channel.cutter_ch == cutter_ch:
                return channel
        return None

    def set_channel(self, channel, sw_open=False):
        """Method to open/close a specific Cutter channel"""
        if not channel in self._channels:
            return False
        status = "open" if sw_open else "close"
        if self._verbose:
            print("Setting cutter {0} channel {1} to {2} ... ". \
                    format(channel.cutter_id, channel.cutter_ch, status), end="")
        #usb and mains cutters have opposite values for ON and OFF
        if channel.cutter_type == "mains":
            sw_open = not sw_open
        self.run(["-as", "{0}".format(channel.cutter_ch),
                  "{0}".format(int(sw_open))], channel.cutter_id)
        if self._verbose:
            print("done.")
        return True

    def close_channel(self, channel):
        """Method to close a specific channel on a cutter"""
        if not self.set_channel(channel, False):
            return False
        time.sleep(self._closing_transient)
        return True

    def close_channels(self, cutter_type="usb"):
        """Method to close all the channels for a certain cutter type"""
        if self._verbose:
            print("Closing all channels of type " + cutter_type + ".")
        self.probe_channels()
        for channel in self._channels:
            if channel.cutter_type == cutter_type:
                self.set_channel(channel, False)
        time.sleep(self._closing_transient)

    def open_channel(self, channel):
        """Method to open a specific channel on a cutter"""
        if not self.set_channel(channel, True):
            return False
        time.sleep(self._opening_transient)
        return True

    def open_channels(self, cutter_type="usb"):
        """Method to open all the channels for a certain cutter type"""
        if self._verbose:
            print("Opening all channels of type " + cutter_type + ".")
        self.probe_channels()
        for channel in self._channels:
            if channel.cutter_type == cutter_type:
                self.set_channel(channel, True)
        time.sleep(self._opening_transient)
