#!/usr/bin/python2
#
# Copyright (c) 2013 Intel, Inc.
# Author igor.stoppa@intel.com
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; version 2 of the License
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the
# Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""Tool for managing collection of devices from the same host PC"""

from __future__ import print_function
import os
import sys
import fcntl
import errno
import time
import atexit
from mechanize import Browser
from urlparse import urlparse
from ConfigParser import ConfigParser
from tizentestrobot.cutter import Cutter
from tizentestrobot.prekit import Prekit
from tizentestrobot.sdb import Sdb
from tizentestrobot.ssh import Ssh
from tizentestrobot.device import Device



VERSION = "0.1.0"

class DevicesManager():
    """Class handling devices connected to the same host PC"""
    def __init__(self):
        self._devices = None
        self._device_lockfile = None
        self.selected_device = None
        self._device_type = None
        self._device_model = None
        self._cutter = Cutter()
        self._prekit = None
        self._sdb = None
        self._ssh = None
        self._username = "ppobs"
        self._password = "Y.508ZadlFQcMXoj"

    def _init_ivis(self):
        """Initializations specific to IVI devices"""
        if not self._ssh:
            self._ssh = Ssh()

    def _init_phones(self):
        """Initializations specific to phone devices"""
        if self._prekit == None and self._sdb == None:
            self._prekit = Prekit()
            self._sdb = Sdb()

    def probe_phones(self):
        """Lists all the visible phones. No change to cutters"""
        self._init_phones()
        return self._sdb.probe_devices() + self._prekit.probe_devices()

    def _ivi_is_ready(self, ivi_ip, mode):
        """Check if the ivi DUT is responsive to ssh"""
        assert(mode in ["SUSE", "Tizen"])
        return False != self._ssh.run(["root@" + ivi_ip,
                              "cat", "/proc/version", "|", "grep", mode])

    def probe_ivis(self):
        """Lists all the visible IVIs. No change to cutters"""
        self._init_ivis()
        leases_file = open("/var/lib/misc/dnsmasq.leases")
        # file format: expiration_time MAC_address IP ....
        leases = leases_file.readlines()
        ids = []
        for i in range(len(leases)):
            leases[i] = leases[i].split()
            if self._ivi_is_ready(leases[i][2], "SUSE"):
                ids.append(leases[i][1])
        return ids

    def detect_phones(self):
        """Lists all the visible phones after closing all cutters"""
        self._cutter.close_channels("usb")
        return self.probe_phones()

    @staticmethod
    def dev_id_to_model(dev_id):
        """Extrapolate the device model from its id"""
        if "CloverTrail" in dev_id:
            return "geek"
        elif "Medfield" in dev_id:
            return "set_blackbay_or_josephine"
        else:
            return "unknown"

    def _parse_filename(self, filename):
        """Extrapolate the device type and model from the name of the image."""
        if "handset-clovertrail" in filename:
            self._device_type = "Phone"
            self._device_model = "geek"
        elif "handset-josephine" in filename:
            self._device_type = "Phone"
            self._device_model = "josephine"
        elif "handset-blackbay" in filename:
            self._device_type = "Phone"
            self._device_model = "blackbay"
        elif "ivi-mbr" in filename:
            self._device_type = "IVI"
            self._device_model = "NDIS"
        elif "generic" in filename and "mbr" in filename:
            self._device_type = "IVI"
            self._device_model = "NDIS"
        else:
            return "unknown"

    def _detect_phones_topology(self):
        """Determine how usb cutters and phones are paired"""
        self._cutter.open_channels("usb")
        self._cutter.close_channels("usb")
        old_dev_ids = self.probe_phones()
        counter = 0
        for channel in self._cutter.probe_channels():
            if channel and channel.cutter_type == "usb":
                self._cutter.open_channel(channel)
            new_dev_ids = self.probe_phones()
            for dev_id in set(old_dev_ids) - set(new_dev_ids):
                name = "Phone_" + "{0:02d}".format(counter)
                model = self.dev_id_to_model(dev_id)
                self._devices.append(Device(name, model, dev_id,
                                     channel.cutter_id, channel.cutter_ch))
                counter += 1
            old_dev_ids = new_dev_ids
        self._cutter.close_channels("usb") #do not leave phones to discharge

    def _detect_ivis_topology(self):
        """Determine how channels of mains cutters and IVIs are paired"""
        counter = 0
        for i in [0, 1]:
            self._cutter.open_channels("mains")
            self._cutter.close_channels("mains")
            print("Waiting for IVI(s) to boot ... ", end="")
            time.sleep(120)
            print("done.")
            old_dev_ids = self.probe_ivis()
            for channel in self._cutter.probe_channels():
                if channel and channel.cutter_type == "mains":
                    self._cutter.open_channel(channel)
                new_dev_ids = self.probe_ivis()
                for dev_id in set(old_dev_ids) - set(new_dev_ids):
                    name = "IVI_" + "{0:02d}".format(counter)
                    model = "NDIS"
                    self._devices.append(Device(name, model, dev_id,
                                         channel.cutter_id, channel.cutter_ch))
                    counter += 1
                old_dev_ids = new_dev_ids

    def detect_topology(self, force=False):
        """Detects topology of connections between host, cutters, DUTs"""
        if self._devices and not force:
            return
        self._devices = []
        self._detect_phones_topology()
        self._detect_ivis_topology()

    def save_topology(self, file_name):
        """Store to a configuration file the layour of DUTs and cutters"""
        config = ConfigParser()
        try:
            config_file = open(file_name, "w")
            for device in self._devices:
                config.add_section(device.name)
                config.set(device.name, "model", device.model)
                config.set(device.name, "id", device.dev_id)
                config.set(device.name, "cutter", device.cutter_id)
                config.set(device.name, "channel", device.cutter_ch)
            config.write(config_file)
            config_file.close()
        except OSError as error:
            sys.exit("Error while saving config file {0}\n {1}". \
                    format(file_name, error))

    def read_topology(self, file_name):
        """Load configuration file with layout of DUTs and cutters"""
        config = ConfigParser()
        self._devices = []
        try:
            config.read(file_name)
            for section in config.sections():
                name = section
                model = config.get(section, "model")
                dev_id = config.get(section, "id")
                cutter_id = config.get(section, "cutter")
                cutter_ch = config.get(section, "channel")
                assert name and model and dev_id and cutter_id
                device = Device(name, model, dev_id, cutter_id, cutter_ch)
                self._devices.append(device)
        except OSError as error:
            sys.exit("Error while loading config file {0}\n {1}". \
                    format(file_name, error))

    def reserve_device(self):
        """Reserves device of specified model from the pool"""
        for device in self._devices:
            if device.model == self._device_model:
                break
        else:
            sys.exit("Error: model not present.")
        while True:
            for device in self._devices:
                if device.model != self._device_model:
                    continue
                print("Attempting to acquire {0} {1} ... ". \
                    format(self._device_model, device.dev_id), end = "")
                try:
                    old_mask = os.umask(011)
                    self._device_lockfile = \
                        os.fdopen(os.open("/var/lock/tztestrobot_" +
                                          device.dev_id,
                                          os.O_WRONLY | os.O_CREAT, 0660),
                                  "w")
                    os.umask(old_mask)
                    fcntl.flock(self._device_lockfile,
                            fcntl.LOCK_EX | fcntl.LOCK_NB)
                    print("success.")
                    self.selected_device = device
                    return device
                except IOError as err:
                    if err.errno in {errno.EACCES, errno.EAGAIN}:
                        print("busy.")
                    else:
                        sys.exit("Error: cannot obtain lock file.")
            print("Sleeping 10s")
            time.sleep(10)
        atexit.register(self._release_device)

    def _release_device(self):
        """Put the reserved device back to the pool."""
        if self.selected_device and self._device_lockfile:
            fcntl.lockf(self._device_lockfile, fcntl.LOCK_UN)
            os.remove("{0}".format(self._device_lockfile.name))
            self._device_lockfile.close()
            self.selected_device = None

    def fetch_image(self, image_url):
        """Downloads remote images"""
        url = urlparse(image_url)
        if url.scheme == "" and url.netloc == "":
            return image_url
        file_name = os.path.basename(url.path)
        browser = Browser()
        if (url.scheme == "https") and \
            (not url.username) and (not url.password) and \
            ("intel.com" in url.netloc.lower()):
            browser.add_password("https://" + url.netloc,
                    self._username, self._password)
        try:
            print("Downloading SW image ... ", end = "")
            browser.retrieve(image_url, file_name)
            print("done.")
        except IOError as error:
            sys.exit("\nError while downloading remote file: " \
                    + image_url + "\n" + str(error))
        return file_name

    def _flash_phone(self, file_name):
        """Method for flashing a phone."""
        if self._sdb.probe_device(self.selected_device.dev_id):
            print("Device in Tizen mode: rebooting to preOS.")
            self._sdb.run(["root", "on"], self.selected_device.dev_id)
            self._sdb.run(["shell", "reboot", "-d", "-f", "-n", "bootloader"],
                    self.selected_device.dev_id)
            time.sleep(15)
        if not self._prekit.probe_device(self.selected_device.dev_id):
            sys.exit("Error: device unreacheable.")
        print("Device in preOS mode: now flashing ...", end = "")
        self._prekit.run([file_name], self.selected_device.dev_id,
                        self._device_model == "blackbay")
        print("done.")
        print("Waiting for reboot in Tizen mode ...", end = "")
        time.sleep(45)
        if not self._sdb.probe_device(self.selected_device.dev_id):
            sys.exit("Error: device unreacheable.")
        print("done.")

    def _reset_selected_ivi(self, ivi_ip, target_mode):
        """Powercycle the ivi that was previously reserved"""
        channel = self._cutter.get_channel(
            self.selected_device.cutter_id,
            self.selected_device.cutter_ch
        )
        if channel == None:
            sys.exit("Bad Channell - cutter_id:{0}  cutter_ch:{1}".\
                format(self.selected_device.cutter_id,
                       self.selected_device.cutter_ch))
        self._cutter.open_channel(channel)
        time.sleep(3)
        self._cutter.close_channel(channel)
        print("Waiting for IVI to boot ...", end="")
        time.sleep(30)
        for i in range(0, 14):
            time.sleep(10)
            print(".", end="")
            if self._ivi_is_ready(ivi_ip, target_mode):
                print(" done.")
                return True
        return False

    def ip_of_selected_ivi(self):
        """Returns IP of selected IVI"""
        leases_file = open("/var/lib/misc/dnsmasq.leases")
        # file format: expiration_time MAC_address IP ....
        lease = None
        for lease in leases_file.readlines():
            lease = lease.split()
            assert(len(lease) == 5)
            if (lease[1] == self.selected_device.dev_id):
                return lease[2]
        return None

    @staticmethod
    def _read_leases():
        """Read from lease file the active leases"""
        with open("/var/lib/misc/dnsmasq.leases") as leases_file:
            return leases_file.readlines()

    def _ip_of_ready_ivi(self):
        """Ensures that IVI has booted from USB stick and returns its IP"""
        lease = None
        leases = self._read_leases()
        if leases == []:
            #No lease found for the device, try to powercycle it blindly
            channel = self._cutter.get_channel(
                          self.selected_device.cutter_id,
                          self.selected_device.cutter_ch)
            self._cutter.open_channel(channel)
            time.sleep(3)
            self._cutter.close_channel(channel)
            time.sleep(120)
            leases = self._read_leases()
        # file format: expiration_time MAC_address IP ....
        for lease in leases:
            lease = lease.split()
            assert(len(lease) == 5)
            if (lease[1] == self.selected_device.dev_id):
                print("Powercycling selected IVI into flashing mode.")
                if not self._reset_selected_ivi(lease[2], "SUSE"):
                    print("IVI not ready: powercycling once more.")
                    if not self._reset_selected_ivi(lease[2], "SUSE"):
                        print("IVI still not ready: last attempt at powercycle.")
                        if not self._reset_selected_ivi(lease[2], "SUSE"):
                            sys.exit("Unable to enter programming mode.")
                break
        assert(len(lease) == 5)
        return lease[2]

    def _prepare_ivi(self, ivi_ip, root_partition):
        """Copy ssh public keys to IVI accounts"""
        print("Copying ssh public key to internal storage ... ", end="")
        if not self._ssh.run(["root@" + ivi_ip,
            "mount", root_partition, "/mnt/sdb_root"]):
            sys.exit("failed mounting internal storage.")
        #do not check for return value: directory might exist
        self._ssh.run(["root@" + ivi_ip,
            "mkdir", "/mnt/sdb_root/root/.ssh"])
        if not self._ssh.run(["root@" + ivi_ip,
                "cat", "/root/.ssh/authorized_keys", ">>",
                "/mnt/sdb_root/root/.ssh/authorized_keys"]):
            sys.exit("failed writing public key.")
        if not self._ssh.run(["root@" + ivi_ip, "sync"]):
            sys.exit("failed flushing internal storage.")
        if not self._ssh.run(["root@" + ivi_ip, "umount", root_partition]):
            sys.exit("failed unmounting internal storage.")
        print("done.")

    def _flash_ivi(self, nfs_filename, ivi_ip):
        """Writes image to internal storage in IVI device"""
        print("Writing image to internal storage ... ", end="")
        assert(self._ssh.run(["root@" + ivi_ip,
                              "bmaptool", "copy", nfs_filename, "/dev/sdb"]))
        print("done.")
        #update info about the partition table
        self._ssh.run(["root@" + ivi_ip, "partprobe", "/dev/sdb"])
        if "generic" in nfs_filename:
            root_partition = "/dev/sdb1"
        else:
            root_partition = "/dev/sdb2"
        self._prepare_ivi(ivi_ip, root_partition)
        if not self._reset_selected_ivi(ivi_ip, "Tizen"):
            sys.exit("IVI failed to boot.")

    def flash_image(self, filename):
        """Validates the image name and tries to flash it"""
        self._parse_filename(filename)
        if self._device_type == "Phone":
            self._init_phones()
            self.reserve_device()
            self._flash_phone(filename)
        elif self._device_type == "IVI":
            # NOTE: it is expected that the image is located somewhere
            # underneath /home and symlinks probably will not work
            if not os.path.isfile(filename):
                sys.exit("Error: {0} not found.".format(filename))
            filename = os.path.abspath(filename)
            # The /home is exported as nfs and mounted remotely as /mnt/nfs
            nfs_filename = filename.replace("/home/", "/mnt/nfs/")
            self._init_ivis()
            self.reserve_device()
            ivi_ip = self._ip_of_ready_ivi()
            self._flash_ivi(nfs_filename, ivi_ip)
        else:
            sys.exit("Can't identify device type/model from filename: {0}".\
                     format(filename))
