#!/usr/bin/python2
#
# Copyright (c) 2013 Intel, Inc.
# Author igor.stoppa@intel.com
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; version 2 of the License
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the
# Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""Class for flashing sw images to selected DUTs from the cfg file"""

from __future__ import print_function
from tizentestrobot.tzflasher import Flasher

import cgi
import collections
import ConfigParser
import datetime
import glob
import os
import signal
import subprocess
import time

VERSION = "0.1.0"

_TEST_MODULES = os.getenv("TZTESTROBOT_MODULES", "/usr/lib/tztestrobot")
_TEST_EXEC_ROOT = os.getenv("TZTESTROBOT_EXECROOT", "./tztestrobot-results.")

# Default timeout (in seconds) for each test script. Can be later
# overridden in <script>.ini file.
_TEST_SCRIPT_TIMEOUT = int(os.getenv("TZTESTROBOT_SCRIPT_TIMEOUT", "300"))

def _find_tests(filename_pattern, executable=True):
    """
    Return list of matching files in installed test modules.
    filename_pattern may contain simple shell-style wildcards (*, ?).
    If executable is True, matching files must be executable.
    """
    print("Building list of tests ... ", end="")
    matching_files = []
    for f in sorted(glob.glob(
            os.path.join(_TEST_MODULES, "*", filename_pattern))):
        if f[-1] == "~": # skip emacs-style backup filenames
            continue
        if executable and not os.access(f, os.X_OK):
            continue
        matching_files.append(f)
    print("{0} found.".format(len(matching_files)))
    return matching_files

_TestResult = collections.namedtuple("TestResult", (
    "module", "name", "passed", "time", "output", "working_dir"))

class _Timeout(Exception):
    pass

def _alarm_handler(signum, frame):
    raise _Timeout

class TzTestRobot(Flasher):
    """Tests SW image"""
    def __init__(self):
        Flasher.__init__(self)

    def _prepare_exec_dir(self, test_filepath):
        """
        Return the name of the directory that has been prepared for executing
        test (test_filepath).
        """
        date = datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")
        try:
            test_module = test_filepath.split("/")[-2]
        except IndexError:
            test_module = ""
        test_file = os.path.basename(test_filepath)
        test_dir = os.path.join(_TEST_EXEC_ROOT + str(os.getpid()),
                                "%s-%s-%s" % (date, test_module, test_file))
        os.makedirs(test_dir)
        os.makedirs(os.path.join(test_dir, "tztestrobot"))
        return test_dir

    def _results_to_xunit(self, start_time, end_time, results):
        """
        Return test results formatted in xunit XML
        """
        xml = [('<?xml version="1.0" encoding="utf-8"?>\n'
               '<testsuite errors="0" failures="%(fails)s" '
               'name="tztestrobot.%(date)s.%(pid)s" skips="0" '
               'tests="%(tests)s" time="%(totaltime)s">\n') % {
                   "fails": len([r for r in results if not r.passed]),
                   "pid": os.getpid(),
                   "date": time.strftime("%Y%m%d%H%M%S",
                                         time.localtime(start_time)),
                   "tests": len(results),
                   "totaltime": "%.3f" % (end_time - start_time)
               }]
        for r in results:
            xml.append(('<testcase classname="%(module)s" name="%(name)s" '
                        'time="%(time)s">\n') % {
                            "module": cgi.escape(r.module, True),
                            "name": cgi.escape(r.name, True),
                            "time": "%.3f" % (r.time,)})
            if not r.passed:
                xml.append(('<failure message="test failure">\n%(output)s\n'
                            '</failure>\n') % {
                                "output": cgi.escape(r.output)
                            })
            xml.append('</testcase>\n')
        xml.append('</testsuite>\n')
        return "".join(xml)

    def _prepare_test_env(self, test_filepath, test_dirpath):
        """
        Return dictionary containing environment for test execution
        """
        test_env = {}
        try:
            ssh_command = (self._devicesmanager._ssh._command + " " +
                str(" ".join(self._devicesmanager._ssh._default_parms)) +
                " root@" + str(self._devicesmanager.ip_of_selected_ivi()))
            test_env["CI_SSH"] = ssh_command
        except AttributeError:
            pass # ssh not available for this platform

        return test_env

    def _run_script(self, script, timeout=None):
        """
        Run executable test in a child process, return _TestResult.
        """
        def record(name, value):
            """save test environment data for debugging"""
            file(os.path.join(working_dir, "tztestrobot", name), "w").write(
                str(value) + "\n")

        script_settings_filename = script + ".ini"
        script_settings = ConfigParser.ConfigParser()
        script_settings.read(script_settings_filename)
        working_dir = self._prepare_exec_dir(script)
        env = self._prepare_test_env(script, working_dir)
        try:
            module_name = script.split("/")[-2]
        except IndexError:
            module_name = "unknown"
        output_filename = os.path.join(working_dir, "tztestrobot/output")

        if timeout == None:
            try:
                timeout = script_settings.getint("Run", "timeout")
                record("timeout", "%s from %s" % (
                    timeout, script_settings_filename))
            except (ConfigParser.NoSectionError, ConfigParser.NoOptionError,
                    ValueError):
                timeout = _TEST_SCRIPT_TIMEOUT
                record("timeout", "%s from env TZTESTROBOT_SCRIPT_TIMEOUT" %
                       (timeout,))
        else:
            record("timeout", "%s from run parameter" % (timeout,))

        record("cmdline", script)
        record("environ", repr(env))
        record("cwd", working_dir)
        output_file = file(output_filename, "w")
        launch_time = time.time()

        signal.signal(signal.SIGALRM, _alarm_handler)
        signal.alarm(timeout)
        print("Running test {0}.".format(script))
        try:
            p = subprocess.Popen([script],
                                 stdin = subprocess.PIPE,
                                 stdout = output_file,
                                 stderr = output_file,
                                 shell = False,
                                 cwd = working_dir,
                                 env = env)
            exit_status = p.wait()
            signal.alarm(0) # reset alarm, do not raise timeout
            record("exitstatus", exit_status)
            print("Test {0} completed with exit status {1}.".\
                  format(script, exit_status))
        except _Timeout:
            exit_status = None
            record("exitstatus", "")
            output_file.write("\nTZTESTROBOT: timeout (%s s)\n" % (timeout,))
            p.terminate()
            print("Test {0} timed out.".format(script))
        total_time = time.time() - launch_time
        record("time", total_time)
        output_file.close()
        result = _TestResult(
            passed = (exit_status == 0),
            time = total_time,
            output = file(output_filename).read(),
            working_dir = working_dir,
            module = module_name,
            name = os.path.basename(script))
        return result

    def run_scripts(self, script_filenames):
        """Execute test scripts in given order"""
        results = []
        start_time = time.time()

        for script in script_filenames:
            results.append(self._run_script(script))
        end_time = time.time()
        if results:
            results_filename = os.path.join(results[0].working_dir,
                                            "..",
                                            "results.xml")
            file(results_filename, "w").write(
                self._results_to_xunit(start_time, end_time, results))
            print("Results saved to {0}.".format(results_filename))

    def smoke_test(self, filenames_and_patterns=("smoke*",)):
        """Execute smoke test modules for DUT"""
        print("Attempting smoke test.")
        script_filenames = []
        for p in filenames_and_patterns:
            script_filenames.extend(_find_tests(p, executable=True))
        return self.run_scripts(script_filenames)
